/* brev - reverse bytes in a file
**
** Copyright (C) 1990 by Jef Poskanzer <jef@mail.acme.com>.
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sysexits.h>

#define INITIAL_CHUNKSIZE 65536

static void reverse_chunk( char* chunk, size_t chunksize );


int
main( int argc, char** argv )
    {
    size_t chunksize;
    char* chunk;
    struct stat sb;
    long off;
    size_t r;
    off_t filesize;

    if ( argc == 2 )
	{
	if ( freopen( argv[1], "r", stdin ) == NULL )
	    {
	    perror( argv[1] );
	    exit( EX_NOINPUT );
	    }
	}
    else if ( argc != 1 )
	{
	(void) fprintf( stderr, "usage:  brev [file]\n" );
	exit( EX_USAGE );
	}

    chunksize = INITIAL_CHUNKSIZE;
    chunk = (char*) malloc( chunksize );
    if ( chunk == (char*) 0 )
	{
	(void) fprintf( stderr, "brev: out of memory\n" );
	exit( EX_OSERR );
	}

    if ( fstat( fileno(stdin), &sb ) == -1 )
	{
	perror( "brev: fstat" );
	exit( EX_IOERR );
	}

    if ( ( sb.st_mode & S_IFMT ) == S_IFIFO )
	{
	/* File is a pipe.  Read in the whole thing. */
	filesize = 0;
	for (;;)
	    {
	    r = fread( &chunk[filesize], 1, chunksize - filesize, stdin );
	    if ( r == 0 )
		break;
	    filesize += r;
	    if ( filesize == chunksize )
		{
		chunksize *= 2;
		chunk = (char*) realloc( chunk, chunksize );
		if ( chunk == (char*) 0 )
		    {
		    (void) fprintf( stderr, "brev: out of memory\n" );
		    exit( EX_OSERR );
		    }
		}
	    }
	reverse_chunk( chunk, filesize );
	}

    else
	{
	/* Not a pipe.  Use seeks and reverse it a chunk at a time. */
	filesize = sb.st_size;
	for ( off = filesize / chunksize * chunksize; off >= 0; off -= chunksize )
	    {
	    if ( fseek( stdin, off, SEEK_SET ) == -1 )
		{
		perror( "brev: fseek" );
		exit( EX_IOERR );
		}
	    r = fread( chunk, 1, chunksize, stdin );
	    if ( r == 0 && ferror( stdin ) )
		{
		perror( "brev: fread" );
		exit( EX_IOERR );
		}
	    reverse_chunk( chunk, r );
	    }
	}

    (void) fclose( stdin );
    (void) fclose( stdout );
    exit( EX_OK );
    }


static void
reverse_chunk( char* chunk, size_t chunksize )
    {
    char* cp1;
    char* cp2;
    char c;

    for ( cp1 = chunk, cp2 = &chunk[chunksize - 1]; cp1 < cp2; ++cp1, --cp2 )
	{
	c = *cp1;
	*cp1 = *cp2;
	*cp2 = c;
	}
    if ( fwrite( chunk, 1, chunksize, stdout ) != chunksize )
	{
	perror( "brev: fwrite" );
	exit( EX_IOERR );
	}
    }
